/* Langton's Ant
 * Author: Aaron Weeden, Shodor, 2016
 */
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

// Declare constants
#define TIME_STEP_COUNT 10
#define ANT_COUNT 10
#define PX_PER_CELL_X 5
#define PX_PER_CELL_Y 5
#define PX_PER_ANT_X 2
#define PX_PER_ANT_Y 2
#define CELL_COUNT_X 21
#define CELL_COUNT_Y 21
#define CELL_COUNT ((CELL_COUNT_X) * (CELL_COUNT_Y))
#define RANDOM_SEED 0
#define PX_PER_GRID_X ((PX_PER_CELL_X) * (CELL_COUNT_X))
#define PX_PER_GRID_Y ((PX_PER_CELL_Y) * (CELL_COUNT_Y))

#define BLACK 0
#define WHITE 1

#define UP    0
#define LEFT  1
#define RIGHT 2
#define DOWN  3

// Map 2D coordinates to 1D coordinates
#define IDX(x, y) ((x) * (CELL_COUNT_X) + (y))

void print(int const timeStep, int cellStates[CELL_COUNT], int const * antDirs,
           int const * antCellXs, int const * antCellYs) {
  printf("Time step %d\n", timeStep);

  // Print cells
  for (int x = 0; x < CELL_COUNT_X; x++) {
    for (int y = 0; y < CELL_COUNT_Y; y++) {
      // Print cell state
      printf("%c", cellStates[IDX(x, y)] == BLACK ? 'B' : '_');

      // Draw ant
      bool isAntHere = false;
      for (int k = 0; k < ANT_COUNT; k++) {
        if (antCellXs[k] == x && antCellYs[k] == y) {
          int antPxX;
          int antPxY;
          switch (antDirs[k]) {
            case UP:
              printf("U");
              antPxX = antCellXs[k] * PX_PER_CELL_X +
                       0.5 * (PX_PER_CELL_X - PX_PER_ANT_X);
              antPxY = antCellYs[k] * PX_PER_CELL_Y;
              break;
            case LEFT:
              printf("L");
              antPxX = antCellXs[k] * PX_PER_CELL_X;
              antPxY = antCellYs[k] * PX_PER_CELL_Y +
                       0.5 * (PX_PER_CELL_Y - PX_PER_ANT_Y);
              break;
            case RIGHT:
              printf("R");
              antPxX = (antCellXs[k] + 1) * PX_PER_CELL_X - PX_PER_ANT_X;
              antPxY = antCellYs[k] * PX_PER_CELL_Y +
                       0.5 * (PX_PER_CELL_Y - PX_PER_ANT_Y);
              break;
            case DOWN:
              printf("D");
              antPxX = antCellXs[k] * PX_PER_CELL_X +
                       0.5 * (PX_PER_CELL_X - PX_PER_ANT_X);
              antPxY = (antCellYs[k] + 1) * PX_PER_CELL_Y - PX_PER_ANT_Y;
              break;
          }
          // Only print direction for one ant
          isAntHere = true;
          break;
        }
      }
      if (!isAntHere) {
        printf(" ");
      }
      printf(" ");
    }
    printf("\n");
  }
  printf("\n");
}

int main() {
  // Initialize cell states
  int cellStates[CELL_COUNT];
  for (int i = 0; i < CELL_COUNT; i++) {
    cellStates[i] = WHITE;
  }

  // For each ant
  int antCellXs[ANT_COUNT];
  int antCellYs[ANT_COUNT];
  int antDirs[ANT_COUNT];
  srandom(RANDOM_SEED);
  for (int i = 0; i < ANT_COUNT; i++) {
    // Give the ant a position
    antCellXs[i] = random() % CELL_COUNT_X;
    antCellYs[i] = random() % CELL_COUNT_Y;

    // Give the ant a direction
    antDirs[i] = LEFT;
  }

  print(0, cellStates, antDirs, antCellXs, antCellYs);

  // Start the simulation
  for (int timeStep = 0; timeStep < TIME_STEP_COUNT; timeStep++) {
    // Turn
    for (int i = 0; i < ANT_COUNT; i++) {
      bool const isBlack = (cellStates[IDX(antCellXs[i],
                                           antCellYs[i])] == BLACK);
      if (     (antDirs[i] == UP    &&  isBlack) ||
               (antDirs[i] == DOWN  && !isBlack)) {
        antDirs[i] = LEFT;
      }
      else if ((antDirs[i] == UP    && !isBlack) ||
               (antDirs[i] == DOWN  &&  isBlack)) {
        antDirs[i] = RIGHT;
      }
      else if ((antDirs[i] == LEFT  &&  isBlack) ||
               (antDirs[i] == RIGHT && !isBlack)) {
        antDirs[i] = DOWN;
      }
      else if ((antDirs[i] == LEFT  && !isBlack) ||
               (antDirs[i] == RIGHT &&  isBlack)) {
        antDirs[i] = UP;
      }
    }

    print(timeStep, cellStates, antDirs, antCellXs, antCellYs);

    // Update cells
    bool shouldFlipCells[CELL_COUNT];
    for (int i = 0; i < CELL_COUNT; i++) {
      shouldFlipCells[i] = false;
    }
    for (int i = 0; i < ANT_COUNT; i++) {
      shouldFlipCells[IDX(antCellXs[i], antCellYs[i])] = true;
    }
    for (int i = 0; i < CELL_COUNT; i++) {
      if (shouldFlipCells[i]) {
        cellStates[i] = (cellStates[i] == BLACK) ? WHITE : BLACK;
      }
    }

    print(timeStep, cellStates, antDirs, antCellXs, antCellYs);

    // Move
    for (int i = 0; i < ANT_COUNT; i++) {
      switch (antDirs[i]) {
        case UP:
          if (antCellYs[i] == 0) {
            antCellYs[i] = CELL_COUNT_Y-1;
          }
          else {
            antCellYs[i]--;
          }
          break;
        case LEFT:
          if (antCellXs[i] == 0) {
            antCellXs[i] = CELL_COUNT_X-1;
          }
          else {
            antCellXs[i]--;
          }
          break;
        case RIGHT:
          if (antCellXs[i] == CELL_COUNT_X-1) {
            antCellXs[i] = 0;
          }
          else {
            antCellXs[i]++;
          }
          break;
        case DOWN:
          if (antCellYs[i] == CELL_COUNT_Y-1) {
            antCellYs[i] = 0;
          }
          else {
            antCellYs[i]++;
          }
          break;
      }
    }

    print(timeStep, cellStates, antDirs, antCellXs, antCellYs);
  } // Continue animation loop

  return 0;
}

